/* ---------------------------------------------------------------- */
//
// MagIO.m
//
// Created by Christoph Birk on 2014-10-03
// Copyright 2014 Carnegie Observatories. All rights reserved.
//
// 2019-05-29  fixed @synchronized() deadlock on open() and close()
// 2022-07-21  added 'dragGprobes' and 'moveOffsetAz'
//
// http://confluence.lco.cl/display/magsoft/Software?preview=/18973023/70451849/TCS_communication.html
//
/* ---------------------------------------------------------------- */

#define DEBUG           0              // debug level
#define TIME_TEST       0

#define RESPONSE_SIZE   512

/* ---------------------------------------------------------------- */

#import  "MagIO.h"
#import  "Logger.h"

#include "utils.h"                     // precess(),walltime()

/* ---------------------------------------------------------------- */

static Logger *magio_logger=nil;       // Class

/* ---------------------------------------------------------------- */

@implementation MagIO
{
  double last_access;
}

/* ---------------------------------------------------------------- */
#pragma mark "Class"
/* ---------------------------------------------------------------- */

+ (void)initialize
{
  extern BOOL use_shared;
  
  if (self != [MagIO class]) return;   // prevent sub-classes
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",__func__,self);
#endif

  magio_logger = [[Logger alloc] initWithBase:"magio" useShared:use_shared];
  magio_logger.mode = LOGGER_FLUSH;    // keep open, auto-flush
}

/* --- */

+ (NSString*)errorString:(int)err
{
  NSString *string=nil;
  
  switch (err) {
    case E_magio_error:   string = @"TelIO: error"; break;
    case E_magio_timeout: string = @"TelIO: timeout"; break;
    case E_magio_reject1: string = @"TelIO: guider probe-1 move rejected"; break;
    case E_magio_reject2: string = @"TelIO: guider probe-2 move rejected"; break;
    case E_magio_reject3: string = @"TelIO: guider probe-3 move rejected"; break;
    default: string = [super errorString:err]; break;
  }
  return string;
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (instancetype)initWithHost:(const char*)hstr port:(int)pint online:(BOOL)on
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s,%d,%d)\n",__func__,hstr,pint,on);
#endif

  self = [super initWithHost:hstr port:pint online:on];

  if (self) {
    last_access = walltime();
    assert(terminator == '\n'); 
  }
  
  return self;
}

/* ---------------------------------------------------------------- */

- (int)open
{
  int err;
#if (DEBUG > 2)
  fprintf(stderr,"%s: count=%d\n",__func__,count);
#endif

  if (![self mylock:lock_timeout]) return E_tcpip_lock; // removed @sync()

  if (count == 0) { char buf[128];
    sprintf(buf,"%s (%d)",__func__,port); [magio_logger append:buf];
  }
  err = [super open];
  if (!err) last_access = walltime();
  [lock unlock];

  return err;
}

/* ---------------------------------------------------------------- */

- (void)close
{
  BOOL r = [self mylock:lock_timeout]; // removed @sync() dead lock
  [super close];
  if (count == 0) { char buf[128];
    sprintf(buf,"%s (%d)",__func__,port); [magio_logger append:buf];
  }
  if (r) [lock unlock];
}

/* ---------------------------------------------------------------- */
#pragma mark "I/O"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (int)request:(const char*)cmd response:(char*)res
{
  char command[256];
#if (DEBUG > 2)
  fprintf(stderr,"%s(%s,%p)\n",__func__,cmd,res);
#endif

  if (![self mylock:lock_timeout]) return E_tcpip_lock;
  [magio_logger append:cmd];
  
  sprintf(command,"%s%c",cmd,terminator);
  int err = [super request:command response:res max:RESPONSE_SIZE];
  last_access = walltime();
  
  if (online) {
    if (err) sprintf(res,"err=%d",err);
    [magio_logger append:res];
  }
  [lock unlock];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)request:(const char*)cmd response:(char*)res timeout:(int)timeout
{
   if (![self mylock:lock_timeout]) return E_tcpip_lock;

   int old_timeout = sock_timeout;
   sock_timeout = timeout;
   int err = [self request:cmd response:res];
   sock_timeout = old_timeout;
  
   [lock unlock];
  
   return err;
}

/* ---------------------------------------------------------------- */

- (int)command:(const char*)cmd
{
  char res[RESPONSE_SIZE];
  
  return [self request:cmd response:res];
}

/* --- */

- (int)command:(const char*)cmd timeout:(int)timeout
{
   if (![self mylock:lock_timeout]) return E_tcpip_lock;

   int old_timeout = sock_timeout;
   sock_timeout = timeout;
   int err = [self command:cmd];
   sock_timeout = old_timeout;
  
   [lock unlock];
  
   return err;
}

/* ---------------------------------------------------------------- */
#pragma mark "Commands"
/* ---------------------------------------------------------------- */

- (int)getAlpha:(double*)alp delta:(double*)del equinox:(double*)equ
{
  double h,m,s;
  char   buf[RESPONSE_SIZE];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%f,%f,%f)\n",__func__,*alp,*del,*equ);
#endif

  int err = [self open]; if (err) return err;
  
  if (!err) {                          // alpha
    err = [self request:"ra" response:buf];
    if (!err && online) {
      sscanf(buf,"%lf:%lf:%lf",&h,&m,&s);
      *alp = h + m/60.0 + s/3600.0;
    }
  }
  if (!err) {                          // delta
    err = [self request:"dec" response:buf];
    if (!err && online) {
      sscanf(buf,"%lf:%lf:%lf",&h,&m,&s);
      *del = fabs(h) + fabs(m)/60.0 + fabs(s)/3600.0;
      if (strchr(buf,'-')) *del *= -1.0;
    }
  }
  if (!err) {                          // equinox
    err = [self request:"epoch" response:buf]; // Note: TCSIS bug (equinox)
    if (!err && online) *equ = atof(buf);
  }
  [self close];

  return err;
}

/* --- */

- (int)getAlpha:(double*)alp delta:(double*)del equinox:(double)equ rot:(float*)rot
{
  double al=*alp,de=*del,eq=equ;
  char   buf[RESPONSE_SIZE];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p,%p,%f,%p)\n",__func__,alp,del,equ,rot);
#endif

  int err = [self open]; if (err) return err;

  if (!err) {
    err = [self getAlpha:&al delta:&de equinox:&eq];
  }
  if (rot) {                           // rotator
    if (!err) err = [self request:"rotangle" response:buf];
    if (!err && online) *rot = (float)atof(buf);
  }

  [self close];

  if (!err) precess(al,de,eq,alp,del,equ);

  return err;
}

/* ---------------------------------------------------------------- */

- (int)moveOffsetRA:(float)ofra dec:(float)ofdc
{
  char cmd[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.1f,%.1f)\n",__func__,ofra,ofdc);
#endif
  
  int err = [self open]; if (err) return err;
  
  // assume guiders are off
  
  sprintf(cmd,"ofra %.2f",ofra);       // offset R.A
  if (!err) err = [self command:cmd];
  sprintf(cmd,"ofdc %.2f",ofdc);       // offset DEC.
  if (!err) err = [self command:cmd];
  if (!err) err = [self command:"offp"];

  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)moveOffsetAz:(float)da el:(float)de
{
  char cmd[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.1f,%.1f)\n",__func__,da,de);
#endif

  sprintf(cmd,"aeg %.2f %.2f",da,de);
  int err = [self command:cmd];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)setAlpha:(double)alpha delta:(double)delta equinox:(double)equinox
{
  char cmd[128],buf[128],*c;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.5f,%.5f,%.2f)\n",__func__,alpha,delta,equinox);
#endif

  int err = [self open]; if (err) return err;

  if (!err) {                          // preset R.A.
    sprintf(cmd,"ra %s",alpha_str(buf,alpha));  
    c = cmd; while (*c) { if (*c == ':') *c = ' '; c++; }
    err = [self command:cmd];
  }
  if (!err) {                          // preset DEC
    sprintf(cmd,"dc %s",delta_str(buf,delta));  
    c = cmd; while (*c) { if (*c == ':') *c = ' '; c++; }
    err = [self command:cmd];
  }
  if (!err) {                          // preset equinox
    sprintf(cmd,"mp %.3f",equinox);      
    err = [self command:cmd];
  }
  if (!err) err = [self command:"opslew "]; // Note: trailing space

  [self close];

  return err;
}

/* --- */

- (int)setAlpha:(double)alpha delta:(double)delta equinox:(double)equinox rot:(float)nro
{
  char cmd[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%.5f,%.5f,%.2f,%.1f)\n",__func__,alpha,delta,equinox,nro);
#endif

  int err = [self open]; if (err) return err;

  if (!err) {                          // preset rotator
    sprintf(cmd,"nro %.1f",nro);
    err = [self command:cmd];
  }
  if (!err) err = [self setAlpha:alpha delta:delta equinox:equinox];

  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)waitMount:(int)coordinated timeout:(int)timeout
{                                      // wait for telescope+guider
  char cmd[128],ret[RESPONSE_SIZE]="000";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%d)\n",__func__,coordinated,timeout);
#endif
#if (TIME_TEST > 0)
  double t1 = walltime();
#endif

  [NSThread sleepForTimeInterval:0.5]; // avoid race condition
  if (coordinated) sprintf(cmd,"gdrmountew %d",timeout);
  else             sprintf(cmd,"mountw %d",timeout);

  int err = [self request:cmd response:ret timeout:(timeout+2)];
  if (!err) { char a,b,c;
    if (coordinated) {
      sscanf(ret,"%c%c%c",&a,&b,&c); a -= '0'; b -= '0'; c -= '0';
      if (b == 4) b = 0;               // NOTE: hack around guider-SW-bug
    } else a = b = c = atoi(ret);
#if (DEBUG > 2)
    fprintf(stderr,"%s(): a=%d, b=%d, c=%d\n",__func__,a,b,c);
#endif
    if ((a == 1) || (b == 1) || (c == 1)) err = E_magio_timeout; // still moving
    if (                         c == 2 ) err = E_magio_reject2; // probe-2 move rejected
    if (             b == 2 )             err = E_magio_reject1; // probe-1 move rejected
    if (            (b == 2) && (c == 2)) err = E_magio_reject3; // both rejected
    if ((a >  1) || (b >  2) || (c >  2)) err = E_magio_error;
  }
  if (!online) [NSThread sleepForTimeInterval:1.0];

#if (TIME_TEST > 0)
  double t2 = walltime();
  sprintf(cmd,"walltime(%s)= %.1f",__func__,t2-t1); [magio_logger append:cmd];
#endif

  return err;
}

/* ---------------------------------------------------------------- */

- (int)setGuiding:(int)g1 g2:(int)g2 timeout:(int)timeout
{ // Note: 'g1/g2' refer to probe-# (not PR/SH designation)
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%d)\n",__func__,g1,g2);
#endif
#if (TIME_TEST > 0)
  char   buf[128];
  double t1 = walltime();
#endif

  int err = [self open]; if (err) return err;

  if (!err) { 
    if (g1) {                          // probe-1: guiding ON
      if (g1 == 2) err = [self command:"gaffiv 0"];  // "F5"
      else         err = [self command:"gafthr 0"];  // "F3"
    } else {
      err = [self command:"gafone 0"];  // "F1"
    }
  }
  
  if (!err) {   
    if (g2) {                          // probe-2: guiding ON
      if (g2 == 2) err = [self command:"gbffiv 0"];  // "F5"
      else         err = [self command:"gbfthr 0"];  // "F3"
    } else {
      err = [self command:"gbfone 0"];  // "F1"
    }
  }

  if (!err) {                          // wait for guiding lock
    if (g1 || g2) (void)[self waitGuiding:timeout];  // Note: ignore error
  }
  
  [self close];

#if (TIME_TEST > 0)
  double t2 = walltime();
  sprintf(buf,"walltime(%s)= %.1f",__func__,t2-t1); [magio_logger append:buf];
#endif

  return err;
}

/* ---------------------------------------------------------------- */

- (int)waitGuiding:(int)timeout        // wait for guiding 'lock'
{
  char cmd[128],ret[RESPONSE_SIZE]="0";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",__func__,timeout);
#endif
#if (TIME_TEST > 0)
  sprintf(cmd,"%s: %d",__func__,timeout); [magio_logger append:cmd];
  double t1 = walltime();
#endif
  
  sprintf(cmd,"gdrlockw %d",timeout);
  int err = [self request:cmd response:ret timeout:(timeout+2)];
  if (!err) {   
    if (atoi(ret) != 0) err = E_magio_timeout;
  }
  if (!online) [NSThread sleepForTimeInterval:1.0];

#if (TIME_TEST > 0)
  double t2 = walltime();
  sprintf(cmd,"walltime(%s)= %.1f",__func__,t2-t1); [magio_logger append:cmd];
#endif

  return err;
}

/* ---------------------------------------------------------------- */

- (int)moveGprobes:(float)dx1 dy1:(float)dy1 dx2:(float)dx2 dy2:(float)dy2
{
  char cmd[128];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.1f,%.1f,%.1f,%.1f)\n",__func__,dx1,dy1,dx2,dy2);
#endif

  int err = [self open]; if (err) return err;

  (void)[self command:"gafone 0"];     // turn-off guiding
  (void)[self command:"gbfone 0"];
  
  if ((dx1 == dx2) && (dy1 == dy2)) {  // both guiders the same motion
    sprintf(cmd,"gprdr3 %.1f %.1f",dx1,dy1);
    if (!err) err = [self command:cmd];
  } else {                             // different motions
    if (dx1 || dy1) { 
      sprintf(cmd,"gprdr1 %.1f %.1f",dx1,dy1);
      if (!err) err = [self command:cmd];
    }
    if (dx2 || dy2) { 
      sprintf(cmd,"gprdr2 %.1f %.1f",dx2,dy2);
      if (!err) err = [self command:cmd];
    }
  }
  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)dragGprobes:(int)mode az:(float)da el:(float)de
{
  char cmd[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d,%.2f,%.2f)\n",__func__,mode,da,de);
#endif

  sprintf(cmd,"gpaer%d %.2f %.2f",mode,da,de);
  int err = [self command:cmd];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)waitGprobes:(int)timeout         
{                                      
  char cmd[128],ret[RESPONSE_SIZE]="000";
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d)\n",__func__,timeout);
#endif
#if (TIME_TEST > 0)
  double t1 = walltime();
#endif

  [NSThread sleepForTimeInterval:0.5];  // avoid race condition

  sprintf(cmd,"gdrew %d",timeout);
  int err = [self request:cmd response:ret timeout:(timeout+2)];
  if (!err) { char a,b,c;
    sscanf(ret,"%c%c%c",&a,&b,&c); a -= '0'; b -= '0'; c -= '0';
    if (b == 4) b = 0;                  // NOTE: hack around guider-SW-bug
#if (DEBUG > 2)
    fprintf(stderr,"%s(): a=%d, b=%d, c=%d\n",__func__,a,b,c);
#endif
    if ((b == 1) || (c == 1)) err = E_magio_timeout;  // still moving
    if (             c == 2 ) err = E_magio_reject2;  // probe-2 move rejected
    if ( b == 2 )             err = E_magio_reject1;  // probe-1 move rejected
    if ((b == 2) && (c == 2)) err = E_magio_reject3;  // both rejected
    if ((b >  2) || (c >  2)) err = E_magio_error;
  }
  if (!online) [NSThread sleepForTimeInterval:2.0];

#if (TIME_TEST > 0)
  double t2 = walltime();
  sprintf(cmd,"walltime(%s)= %.1f",__func__,t2-t1); [magio_logger append:cmd];
#endif

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getGprobe:(int)gp f:(float*)focus x:(float*)xpos y:(float*)ypos
{
  char cmd[64],buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%p,%p,%p)\n",__func__,gp,focus,xpos,ypos);
#endif

  int err = [self open];  if (err) return err;

  if (!err && focus) {
    sprintf(cmd,"guiderf%d",gp);
    err = [self request:cmd response:buf];
    if (!err && online) *focus = (float)atof(buf);
  }
  if (!err && xpos) {
    sprintf(cmd,"guiderx%d",gp);
    err = [self request:cmd response:buf];
    if (!err && online) *xpos  = (float)atof(buf);
  }
  if (!err && ypos) {              
    sprintf(cmd,"guidery%d",gp);
    err = [self request:cmd response:buf];
    if (!err && online) *ypos  = (float)atof(buf);
  }
#if (DEBUG > 1)
  if (!err) {
    sprintf(cmd,"p%dmask",gp);
    err = [self request:cmd response:buf]; // IDEA: analyze result
  }
#endif

  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)setGprobe:(int)gp focus:(float)focus
{
  char cmd[64];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%d,%.2f)\n",__func__,gp,focus);
#endif

  sprintf(cmd,"gpF%dA %.2f",gp,focus);

  return [self command:cmd];
}

/* ---------------------------------------------------------------- */

- (int)getUT:(int*)ut st:(int*)st
{  
  int  hr,mi,sc;
  char buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p)\n",__func__,ut,st);
#endif

  int err = [self open];  if (err) return err;

  if (!err && ut) {                    // get universal time
    err = [self request:"ut" response:buf];
    if (!err && online) {
      (void)sscanf(buf,"%d:%d:%d",&hr,&mi,&sc);
      *ut = hr*3600 + mi*60 + sc;
    }
  }
  if (!err && st) {                    // get siderial time
    err = [self request:"st" response:buf];
    if (!err && online) {
      (void)sscanf(buf,"%d:%d:%d",&hr,&mi,&sc);
      *st = hr*3600 + mi*60 + sc;
    }
  }
  [self close];

  return err;
}


/* ---------------------------------------------------------------- */

- (int)getWXT:(float*)temp p:(float*)pres h:(float*)humi w:(float*)wind d:(float*)wdir
{
  char buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%p,%p)\n",__func__,temp,pres,humi,wind);
#endif

  int err = [self open];  if (err) return err;

  if (!err && temp) {                // IDEA: "telenv" ?
    err = [self request:"wxtemp" response:buf];
    if (!err && online) *temp = (float)atof(buf);
  }
  if (!err && pres) {
    err = [self request:"wxpres" response:buf];
    if (!err && online) *pres = (float)atof(buf);
  }
  if (!err && humi) {
    err = [self request:"wxhumid" response:buf];
    if (!err && online) *humi = (float)atof(buf);
  }
  if (!err && wind) {
    err = [self request:"wxwind" response:buf];
    if (!err && online) *wind = (float)atof(buf);
  }
  if (!err && wdir) {
    err = [self request:"wxwdir" response:buf];
    if (!err && online) *wdir = (float)atof(buf);
  }
  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getAmbient:(float*)ta cell:(float*)tc truss:(float*)tt m2:(float*)t2
{
  char buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%p)\n",__func__,ta,tc,tt);
#endif

  int err = [self open]; if (err) return err;

  if (ta) {                       // IDEA: "telenv" ?
    if (!err) err = [self request:"tambient" response:buf];
    if (!err && online) *ta = (float)atof(buf);
  }
  if (tc) {
    if (!err) err = [self request:"tcell" response:buf];
    if (!err && online) *tc = (float)atof(buf);
  }
  if (tt) {
    if (!err) err = [self request:"ttruss" response:buf];
    if (!err && online) *tt = (float)atof(buf);
  }
  if (t2) {
    if (!err) err = [self request:"tseccel" response:buf];
    if (!err && online) *t2 = (float)atof(buf);
  }
  [self close];
  
  return err;
}

/* ---------------------------------------------------------------- */

- (int)getAirmass:(float*)am rotangle:(float*)ra rotatore:(float*)re
          azimuth:(float*)az elevation:(float*)el para:(float*)pa
{
  char buf[RESPONSE_SIZE];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p,%p,%p,%p,%p)\n",__func__,am,ra,re,az,el,pa);
#endif

  int err = [self open]; if (err) return err;

  if (am) {
    if (!err) err = [self request:"airmass" response:buf];
    if (!err && online) *am = (float)atof(buf);
  }
  if (ra) {
    if (!err) err = [self request:"rotangle" response:buf];
    if (!err && online) *ra = (float)atof(buf);
  }
  if (re) {
    if (!err) err = [self request:"rotatore" response:buf];
    if (!err && online) *re = (float)atof(buf);
  }
  if (az) {
    if (!err) err = [self request:"telaz" response:buf];
    if (!err && online) *az = (float)atof(buf);
  }
  if (el) {
    if (!err) err = [self request:"telel" response:buf];
    if (!err && online) *el = (float)atof(buf);
  }
  if (pa) {
    if (!err) err = [self request:"telpa" response:buf];
    if (!err && online) *pa = (float)atof(buf);
  }
  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

- (int)getCatalog:(char*)name
{
  char buf[RESPONSE_SIZE];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",__func__,name);
#endif

  int err = [self open]; if (err) return err;

  if (name) {
    if (!err) err = [self request:"catobj" response:buf];
    if (!err && online) strcpy(name,buf);
  }

  [self close];

  return err;
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
