/* ---------------------------------------------------------------- */
//
//  AppDelegate.m
//  GuidePaddle
//
//  Created by Christoph Birk on 2022-06-23
//  Copyright © 2022 Carnegie Observatories. All rights reserved.
//
/* ---------------------------------------------------------------- */

#define DEBUG     1

/* ---------------------------------------------------------------- */

#import "AppDelegate.h"
#import "main.h"
#import "MagIO.h"

#include "utils.h"

/* ---------------------------------------------------------------- */

#define DBE_CONFIG      "dbe_configuration"
#define DBE_TCS_HOST    "dbe_tcs_host"
#define DBE_TCS_PORT    "dbe_tcs_port"
#define DBE_OFFSET      "dbe_offset"
#define DBE_ELSIGN      "dbe_elsign"
#define DBE_PARITY      "dbe_parity"      // b0008
#define DBE_FRAME       "dbe_windowFrame"

/* ---------------------------------------------------------------- */

@interface AppDelegate ()

@property (weak) IBOutlet NSWindow *window;
@end

/* ---------------------------------------------------------------- */

@implementation AppDelegate
{
  char host[128];
  int  port;
  double pa,elev,para,offset,elsign,elevation,parity;
  MagIO *tcs;
  NSArray *instruments;
  BOOL prefOpen;
}

/* ---------------------------------------------------------------- */
#pragma mark "NSApp Delegate"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)applicationDidFinishLaunching:(NSNotification*)aNotification
{
  NSString *str;
  char buf[512];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",__func__,aNotification);
#endif

  YbundleInfo(NULL,NULL,NULL,NULL,&r_path,NULL);

  // defaults write <app domain name> NSAppSleepDisabled -bool YES
  if (!YappNapDisabled()) {
    [Notice openNotice:"Warning: AppNap enabled\rPlease restart the GUI"
            title:P_TITLE time:0 level:1];
  }
  assert(MAC_OS_X_VERSION_MIN_REQUIRED > MAC_OS_X_VERSION_10_9);

  NSArray *users = [NSArray arrayWithObjects:@"birk",nil];
  if ([users indexOfObject:NSUserName()] == NSNotFound) {
    use_shared = YprefSetupShared();
  } else {
    use_shared = NO;
  }
  (void)YappSupportCreate(&s_path,use_shared);

  main_logger = [[Logger alloc] initWithBase:nil useShared:use_shared];
  str = [NSString stringWithFormat:@"user= %@ , sharedSetup= %d , UT= %s",
           NSUserName(),use_shared,get_ut_timestr(buf,cor_time(0))];
  [main_logger append:[str UTF8String]];
  
  sprintf(buf,"%s/instruments.xml",s_path); // in ApplicationSupport
  if (!YfileReadable(buf)) { NSString *src; // copy from Resources
    src = [[NSBundle mainBundle] pathForResource:@"instruments" ofType:@"xml"];
    YfileCopy([src UTF8String],buf);
    YfilePermissions(buf,6,6,6);  // rw-,rw-,rw-
  }
  instruments = YplistWithContentsOfFile(@(buf));
  for (NSDictionary *d in instruments) { // setup instrument popup
    NSString *name = [d objectForKey:@"name"];
    [popup_inst addItemWithTitle:name];
  }
  [instruments retain];

  [self setupConfig];
  
  YwindowFrameSet(gpWindow,DBE_FRAME);
}

/* ---------------------------------------------------------------- */

- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication*)appl
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",__func__,appl);
#endif

  return YES;
}

/* ---------------------------------------------------------------- */

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication*)appl
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",__func__,appl);
#endif

  return NSTerminateNow;
}

/* ---------------------------------------------------------------- */

- (void)applicationWillTerminate:(NSNotification*)aNotification
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",__func__,aNotification);
#endif

}

/* ---------------------------------------------------------------- */

- (void)windowWillClose:(NSNotification*)note
{
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,note);
#endif

  if (note.object == self.window) {
    prefOpen = FALSE;
  } else
  if (note.object == gpWindow) {
    YwindowFrameSave(gpWindow,DBE_FRAME);
  }
}

/* ---------------------------------------------------------------- */
#pragma mark "Instance"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (void)setupConfig
{
  int i = YprefGetInt(DBE_CONFIG,0,use_shared);
  [popup_inst selectItemAtIndex:i];
  
  if (i > 0) {
    NSDictionary *d = instruments[i-1];
    strcpy(host,[[d objectForKey:@"tcsHost"] UTF8String]);
    port = [[d objectForKey:@"tcsPort"] intValue];
    offset = [[d objectForKey:@"offset"] doubleValue];
    elsign = [[d objectForKey:@"elevationSign"] doubleValue];
    parity = [[d objectForKey:@"parity"] doubleValue];
  } else {
    YprefGetChar(DBE_TCS_HOST,"localhost",host,sizeof(host),use_shared);
    port = YprefGetInt(DBE_TCS_PORT,5801,use_shared);
    offset = YprefGetDouble(DBE_OFFSET,-120.0,use_shared);
    elsign = YprefGetDouble(DBE_ELSIGN,-1.0,use_shared);
    parity = YprefGetDouble(DBE_PARITY,+1.0,use_shared);
  }
  [edit_tcs_host setStringValue:@(host)];
  [edit_tcs_port setStringValue:[NSString stringWithFormat:@"%d",port]];
  [edit_offset setStringValue:[NSString stringWithFormat:@"%.1f",offset]];
  [popup_elsign selectItemAtIndex:((elsign < 0) ? 0 : 1)];
  [popup_parity selectItemAtIndex:((parity < 0) ? 0 : 1)];

  YcontrolEnable(edit_tcs_host,!i);
  YcontrolEnable(edit_tcs_port,!i);
  YcontrolEnable(edit_offset,!i);
  YcontrolEnable(popup_elsign,!i);
  YcontrolEnable(popup_parity,!i);
}

/* ---------------------------------------------------------------- */

- (void)drawCompass:(NSImageView*)view color:(NSColor*)color north:(double)north east:(double)east
{
  CGContextRef context;
  NSRect       bounds;
  NSImage      *image;
  float        r,x,y,xc,yc;
  CGFloat      re,gr,bl,al;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%.1f,%.1f)\n",PREFUN,north,east);
#endif
  assert([NSThread isMainThread]);

  bounds = [view bounds];
  image  = [[NSImage alloc] initWithSize:NSMakeSize(bounds.size.width,bounds.size.height)];
  [image lockFocus];
  
  context = [[NSGraphicsContext currentContext] CGContext];
  xc = bounds.size.width/2; yc = bounds.size.height/2; r = 0.8f*fmin(xc,yc);
  CGContextBeginPath(context);
  CGContextSetLineWidth(context,1);
  CGContextSetRGBStrokeColor(context,0.0,0.0,0.0,1.0); // black
  CGContextAddArc(context,xc,yc,r,0.0f,2*3.14159,0);
  CGContextDrawPath(context,kCGPathStroke);
  [color getRed:&re green:&gr blue:&bl alpha:&al];
  CGContextSetRGBStrokeColor(context,re,gr,bl,al);
  CGContextBeginPath(context);
  CGContextSetLineWidth(context,2);
  CGContextMoveToPoint(context,xc,yc);
  x = xc + r*sin(north/RADS);
  y = yc + r*cos(north/RADS);  // Note: inverted sign as X11
  CGContextAddLineToPoint(context,x,y);
  CGContextMoveToPoint(context,xc,yc);
  x = xc + (r/2.0)*sin(east/RADS);
  y = yc + (r/2.0)*cos(east/RADS);
  CGContextAddLineToPoint(context,x,y);
  CGContextDrawPath(context,kCGPathStroke);

  [image unlockFocus];
  [view setImage:image];
  [image release];
}

/* --- */

- (void)updateCompass
{
  float north,east;
#if (DEBUG > 0)
  fprintf(stderr,"%s: %.1f %.1f\n",PREFUN,pa,para);
#endif
  assert([NSThread isMainThread]);

  north =  parity*(-fabs(pa) + para);    // b0008
  if (parity > 0) east = north + ((pa > 0) ? +90.0f : -90.0f); // b0008
  else            east = north + ((pa < 0) ? +90.0f : -90.0f);
  [self drawCompass:view_compass color:[NSColor greenColor] north:north east:east];

  north =  -fabs(pa)*parity;
  if (parity > 0) east = north + ((pa < 0) ? +90.0f : -90.0f);
  else            east = north + ((pa > 0) ? +90.0f : -90.0f);
  [self drawCompass:view_tele color:[NSColor redColor] north:north east:east];
}

/* ---------------------------------------------------------------- */
#pragma mark "Actions"
/* ---------------------------------------------------------------- */
//34567890123456789012345678901234567890

- (IBAction)action_config:(id)sender
{
  char buf[256];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (sender == popup_inst) {
    int i = (int)[popup_inst indexOfSelectedItem];
    YprefPutInt(DBE_CONFIG,i,use_shared);
    [self setupConfig];
  } else
  if ([[sender title] compare:@"Pref" options:0 range:NSMakeRange(0,4)] == NSOrderedSame) {
    [self.window makeKeyAndOrderFront:self];
    prefOpen = TRUE;             // closes open TCS connetions
  } else
  if ([[sender title] compare:@"OK"] == NSOrderedSame) {
    if ([popup_inst indexOfSelectedItem] == 0) { // Other
      YeditGet(edit_tcs_host,host,sizeof(host));
      YprefPutChar(DBE_TCS_HOST,host,use_shared);
      port = [edit_tcs_port intValue];
      YprefPutInt(DBE_TCS_PORT,port,use_shared);
      offset = [edit_offset doubleValue];
      YprefPutDouble(DBE_OFFSET,offset,use_shared);
      elsign = ([popup_elsign indexOfSelectedItem]) ? 1.0 : -1.0;
      YprefPutDouble(DBE_ELSIGN,elsign,use_shared);
      parity = ([popup_parity indexOfSelectedItem]) ? 1.0 : -1.0;
      YprefPutDouble(DBE_PARITY,parity,use_shared);
    }
    sprintf(buf,"TCSIS= %s:%d",host,port);
    [main_logger append:buf];
    [gpWindow setTitle:[NSString stringWithFormat:@"%s/%@",P_TITLE,[popup_inst title]]];
    [gpWindow makeKeyAndOrderFront:self];
    if (tcs) { [tcs release]; tcs = nil; }
    prefOpen = FALSE;
    if (port > 0) tcs = [[MagIO alloc] initWithHost:host port:port online:YES];
    [self.window orderOut:self];
    [self action_update:nil];
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_update:(id)sender
{
  int err=0;
  double f=0;
  float el,pr;
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

#if 0 // TESTING
  el = [edit_elev doubleValue];
  pr = [edit_para doubleValue];
#else
  err = [tcs getAirmass:NULL rotangle:NULL rotatore:NULL azimuth:NULL elevation:&el para:&pr];
  if (err) { char buf[128];
    sprintf(buf,"failed to open connection to TCSIS (err=%d)",err);
    [main_logger message:buf level:1 time:30 file:YES];
    return;
  }
#endif
  if (!err) {
    f = offset + elsign*el;
    if (f == 0.0) f = 360.0;
    while (f >  360.0) f -= 360.0;
    while (f < -360.0) f += 360.0;
    pa = f;
    elev = el;
    para = pr;
    [edit_elev setIntegerValue:(int)my_round(elev,0)];
    [edit_para setIntegerValue:(int)my_round(para,0)];
  }
  [edit_pa setIntegerValue:(int)my_round(pa,0)];
  [self updateCompass];
}

/* ---------------------------------------------------------------- */

- (IBAction)action_move:(id)sender
{
  double dx=0,dy=0,da,de,R=0.1; // TODO variable distance
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  int err = [tcs open];
  if (err) { char buf[128];
    sprintf(buf,"failed to open connection to TCSIS (err=%d)",err);
    [main_logger message:buf level:1 time:15 file:YES];
    return;
  }
  [self action_update:sender];
  
  if ([[sender title] compare:@"Up"] == NSOrderedSame)    dy = -R;
  if ([[sender title] compare:@"Down"] == NSOrderedSame)  dy = +R;
  if ([[sender title] compare:@"Left"] == NSOrderedSame)  dx = -parity*R; // b0007
  if ([[sender title] compare:@"Right"] == NSOrderedSame) dx =  parity*R; // b0008
  rotate(dx,dy,pa,&da,&de);
  // printf("dx=%.2f, dy=%.2f, da=%.2f, de=%.2f\n",dx,dy,da,de);

  dispatch_async(GLOBAL_QUEUE,^(void) {
    [tcs dragGprobes:3 az:-da el:-de];
    [tcs moveOffsetAz:da el:de];
    NSDate *date = [NSDate date];
    while (YsecSince(date) < 5.0) { Ysleep(350); if (prefOpen) break; }
    [tcs close];
  });

}

/* ---------------------------------------------------------------- */

- (IBAction)action_file:(id)sender
{
  char title[64];
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  Ystr2chr([sender title],title,sizeof(title));
  if (strstr(title,"Logfile")) {     // open logfile
#if (MAC_OS_X_VERSION_MIN_REQUIRED > MAC_OS_X_VERSION_10_13)
    [[NSWorkspace sharedWorkspace] openURL:[NSURL fileURLWithPath:main_logger.logfile]];
#else
    [[NSWorkspace sharedWorkspace] openFile:main_logger.logfile];
#endif
  } else
  if (strstr(title,"AppSupport")) {  // Application Support
    NSURL *url = [NSURL fileURLWithPath:@(s_path) isDirectory:YES];
    [[NSWorkspace sharedWorkspace] openURL:url];
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

- (IBAction)action_help:(id)sender   // b0008
{
#if (DEBUG > 0)
  fprintf(stderr,"%s:%p\n",PREFUN,sender);
#endif
  assert([NSThread isMainThread]);

  if (strstr([[sender title] UTF8String],"Paddle")) {
    Ywww("http://instrumentation.obs.carnegiescience.edu/Software/GuidePaddle");
  } else {
    assert(0);
  }
}

/* ---------------------------------------------------------------- */

@end

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
